# # -*- coding: utf-8 -*-
# import os
# from dataclasses import dataclass, field
# import numpy as np
# import pandas as pd
# from typing import List, Tuple, Union
# import datetime
# from frm.enums.utils import RollConv, TimingConvention, Stub, PeriodFreq, DayRoll, DayCountBasis, ExchangeNotionals
# from frm.utils.daycount import year_frac, day_count
#
#
# if __name__ == "__main__":
#     os.chdir(os.environ.get('PROJECT_DIR_FRM'))
#
# # def set_default(value, default):
# #     if pd.isna(value) or value is None:
# #         return default
# #     return value
#
#
# @dataclass
# class Schedule:
#     # Schedule parameters
#     start_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime]
#     end_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime]
#     freq: PeriodFreq
#     roll_conv: RollConv = RollConv.MODIFIED_FOLLOWING
#     day_roll: DayRoll = DayRoll.UNADJUSTED
#     first_period_end: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None
#     last_period_start: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None
#     first_stub: Stub = Stub.DEFAULT
#     last_stub: Stub = Stub.DEFAULT
#     roll_user_specified_dates: bool = False
#     cal: np.busdaycalendar = np.busdaycalendar()
#     # Coupon payment date parameters
#     add_cpn_payment_dates: bool = False
#     cpn_payment_delay: int = 0
#     cpn_payment_timing: TimingConvention = TimingConvention.IN_ARREARS
#     # Notional schedule parameters
#     add_notional_schedule: bool = False
#     notional_amount: float | np.ndarray = 100_000_000
#     exchange_notionals: ExchangeNotionals = ExchangeNotionals.NEITHER
#     add_notional_payment_dates: bool = False
#     notional_payment_delay: int = 0
#     notional_payment_timing: TimingConvention = TimingConvention.IN_ARREARS
#
#     # Fixing date parameters
#     fixing_data_params = None
#     #add_fixing_dates: bool = False
#     #fixing_days_ahead: int = 0
#     #fixing_timing: TimingConvention = TimingConvention.IN_ADVANCE
#
#     # The schedule DataFrame is initialized after the object is created
#     df: pd.DataFrame = field(init=False)
#
#     def __post_init__(self):
#         # Convert to pandas Timestamp
#         self.start_date = pd.Timestamp(self.start_date)
#         self.end_date = pd.Timestamp(self.end_date)
#         self.first_period_end = pd.Timestamp(self.first_period_end) if self.first_period_end is not None else None
#         self.last_period_start = pd.Timestamp(self.last_period_start) if self.last_period_start is not None else None
#
#         self.generate_schedule()
#
#         if self.add_cpn_payment_dates:
#             self.add_payment_dates_to_schedule(self.cpn_payment_delay, self.cpn_payment_timing, prefix='cpn_')
#
#         if self.add_notional_schedule:
#             self.setup_notional_schedule(self.notional_amount, self.exchange_notionals)
#
#     def generate_schedule(self):
#         # Generate the schedule DataFrame using the stored parameters
#         self.df = make_schedule(
#             start_date=self.start_date,
#             end_date=self.end_date,
#             freq=self.freq,
#             roll_conv=self.roll_conv,
#             day_roll=self.day_roll,
#             first_period_end=self.first_period_end,
#             last_period_start=self.last_period_start,
#             first_stub=self.first_stub,
#             last_stub=self.last_stub,
#             roll_user_specified_dates=self.roll_user_specified_dates,
#             cal=self.cal
#         )
#
#         if self.roll_user_specified_dates:
#             self.start_date = self.df['period_start'].iloc[0]
#             self.end_date = self.df['period_end'].iloc[-1]
#
#
#     def update_and_regenerate(self, **kwargs):
#         """
#         Update specified parameters and regenerate the schedule.
#
#         Parameters:
#             **kwargs: Dictionary of parameters to update.
#         """
#         # Update the parameters
#         for key, value in kwargs.items():
#             if hasattr(self, key):
#                 setattr(self, key, value)
#             else:
#                 raise AttributeError(f"'Schedule' object has no attribute '{key}'")
#         # Regenerate the schedule with updated parameters
#         self.generate_schedule()
#
#
#     def add_payment_dates_to_schedule(self,
#                                       payment_delay: int=0,
#                                       payment_timing: TimingConvention = TimingConvention.IN_ARREARS,
#                                       prefix: str=''):
#         """
#         Add payment dates to the schedule DataFrame.
#
#         Parameters
#         ----------
#         payment_delay : int
#             Specifies how many days after period start_date/end_date (if payments are in_advance/in_arrears), the payment is made. The default is 0.
#         payment_timing : TimingConvention
#             Specifies when payments are made. The default is TimingConvention.IN_ARREARS.
#         prefix: str
#             Prefix to add to the payment_date column name. The default is ''.
#         """
#         match payment_timing:
#             case TimingConvention.IN_ARREARS:
#                 dates = self.df['period_end'].to_numpy(dtype='datetime64[D]')
#             case TimingConvention.IN_ADVANCE:
#                 dates = self.df['period_start'].to_numpy(dtype='datetime64[D]')
#             case _:
#                 raise ValueError(f"Invalid payment_timing {payment_timing}")
#
#         dates_np = np.busday_offset(dates, offsets=payment_delay, roll=self.roll_conv.value, busdaycal=self.cal)
#         self.df[prefix + 'payment_date'] = pd.DatetimeIndex(dates_np).astype('datetime64[ns]')
#
#
#     def add_fixing_dates_to_schedule(self,
#                                      fixing_days_ahead: int=0,
#                                      fixing_timing: TimingConvention = TimingConvention.IN_ADVANCE,
#                                      roll_conv: RollConv=None):
#
#         if roll_conv is None:
#             roll_conv = self.roll_conv
#
#         match fixing_timing:
#             case TimingConvention.IN_ARREARS:
#                 dates = self.df['period_end'].to_numpy(dtype='datetime64[D]')
#             case TimingConvention.IN_ADVANCE:
#                 dates = self.df['period_start'].to_numpy(dtype='datetime64[D]')
#             case _:
#                 raise ValueError(f"Invalid fixing_timing {fixing_timing}")
#
#         dates_np = np.busday_offset(dates, offsets=-1 * fixing_days_ahead, roll=roll_conv.value, busdaycal=self.cal)
#         self.df.instert(0, 'fixing_date', pd.DatetimeIndex(dates_np).astype('datetime64[ns]'))
#
#
#     def setup_notional_schedule(self,
#                                 notional_amount: float | np.ndarray,
#                                 exchange_notionals: ExchangeNotionals=ExchangeNotionals.NEITHER,
#                                 initial_notional_exchange_date=None):
#
#         notional_amount = np.atleast_1d(notional_amount)
#         if notional_amount.shape == (1,):
#             self.df['notional_payment'] = 0
#             self.df['notional'] = notional_amount[0]
#         elif notional_amount.shape == self.df.shape[0]:
#             self.df['notional'] = notional_amount
#             self.df['notional_payment'] = notional_amount[:-1] - notional_amount[1:]
#         else:
#             raise ValueError("Invalid notional_amount shape")
#
#         self.add_payment_dates_to_schedule(self.notional_payment_delay, self.notional_payment_timing,
#                                            prefix='notional_')
#
#         if exchange_notionals == ExchangeNotionals.START or exchange_notionals == ExchangeNotionals.BOTH:
#             # Add initial notional exchange period
#             self._add_initial_notional_exchange_period_to_schedule(initial_notional_exchange_date)
#             self.df.loc[self.df.index[0], 'notional'] = 0
#             self.df.loc[self.df.index[0], 'notional_payment'] = -1 * notional_amount[0]
#
#         if exchange_notionals == ExchangeNotionals.END or exchange_notionals == ExchangeNotionals.BOTH:
#             # Add final notional exchange period
#             self.df.loc[self.df.index[-1], 'notional_payment'] = notional_amount[-1]
#
#
#     def _add_initial_notional_exchange_period_to_schedule(self, initial_notional_exchange_date=None):
#         #assert 'period_years' in self.df.columns
#         assert 'notional_payment_date' in self.df.columns
#         column_order = self.df.columns
#
#         if initial_notional_exchange_date is None:
#             initial_notional_exchange_date = self.df['period_start'].iloc[0]
#
#         row_data = {'notional_payment_date': initial_notional_exchange_date}
#         self.df = pd.concat([pd.DataFrame(row_data, index=[0]), self.df], ignore_index=True)
#
#         self.df = self.df[column_order]
#         self.df.reset_index(drop=True, inplace=True)
#
#
#     def add_period_daycount(self, day_count_basis: DayCountBasis):
#         """ Add the period length in days and years to the schedule DataFrame, always replacing existing columns"""
#
#         # Remove 'period_days' and 'period_years' if they already exist
#         if 'period_daycount' in self.df.columns:
#             self.df.pop('period_daycount')
#
#         # Get the column index for 'period_end'
#         col_index = self.df.columns.get_loc('period_end')
#
#         # Insert 'period_days' and 'period_years' at the specified positions
#         mask = np.logical_and(self.df['period_start'].notnull(), self.df['period_end'].notnull())
#         days = np.full(self.df.shape[0], np.nan)
#         days[mask] = day_count(self.df['period_start'][mask], self.df['period_end'][mask], day_count_basis)
#         self.df.insert(loc=col_index + 1, column='period_daycount',value=days)
#
#     def add_period_yearfrac(self, day_count_basis: DayCountBasis):
#         """ Add the period length in days and years to the schedule DataFrame, always replacing existing columns"""
#
#         if 'period_yearfrac' in self.df.columns:
#             self.df.pop('period_yearfrac')
#
#         # Get the column index for 'period_end'
#         col_index = self.df.columns.get_loc('period_end')
#
#         # Insert 'period_days' and 'period_years' at the specified positions
#         mask = np.logical_and(self.df['period_start'].notnull(), self.df['period_end'].notnull())
#         years = np.full(self.df.shape[0], np.nan)
#         years[mask] = year_frac(self.df['period_start'][mask], self.df['period_end'][mask], day_count_basis)
#         self.df.insert(loc=col_index + 2, column='period_yearfrac', value=years)
#
#     def copy(self):
#         # Create a new instance of Schedule - refresh with ChatGPT if main class changes
#         new_schedule = Schedule(
#             start_date=self.start_date,
#             end_date=self.end_date,
#             freq=self.freq,
#             roll_conv=self.roll_conv,
#             day_roll=self.day_roll,
#             first_period_end=self.first_period_end,
#             last_period_start=self.last_period_start,
#             first_stub=self.first_stub,
#             last_stub=self.last_stub,
#             roll_user_specified_dates=self.roll_user_specified_dates,
#             cal=np.busdaycalendar(weekmask=self.cal.weekmask, holidays=self.cal.holidays),
#             add_cpn_payment_dates=self.add_cpn_payment_dates,
#             cpn_payment_delay=self.cpn_payment_delay,
#             cpn_payment_timing=self.cpn_payment_timing,
#             add_notional_schedule=self.add_notional_schedule,
#             notional_amount=self.notional_amount,
#             exchange_notionals=self.exchange_notionals,
#             add_notional_payment_dates=self.add_notional_payment_dates,
#             notional_payment_delay=self.notional_payment_delay,
#             notional_payment_timing=self.notional_payment_timing,
#         )
#
#         # Deep copy of the schedule DataFrame
#         new_schedule.df = self.df.copy(deep=True)
#
#         return new_schedule
#
#
# def variable_notional_helper():
#     # TODO
#     #  1. Increasing or decreasing amount or percentage
#     #  2. Annuity (custom fixed rate or floating rate) in order to get fixed total payments (coupon + notional) over the life
#     pass
#
#
# def make_schedule(
#         start_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime],
#         end_date: [pd.Timestamp, np.datetime64, datetime.date, datetime.datetime],
#         freq: PeriodFreq,
#         roll_conv: RollConv=RollConv.MODIFIED_FOLLOWING,
#         day_roll: DayRoll=DayRoll.UNADJUSTED,
#         first_period_end: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None,
#         last_period_start: Union[pd.Timestamp, np.datetime64, datetime.date, datetime.datetime, None] = None,
#         first_stub: Stub=Stub.DEFAULT,
#         last_stub: Stub=Stub.DEFAULT,
#         roll_user_specified_dates: bool=False,
#         cal: np.busdaycalendar=np.busdaycalendar(),
#         ) -> pd.DataFrame:
#     """
#     Create a schedule. Optional detailed stub logic.
#
#     Parameters
#     ----------
#     start_dt : pandas.Timestamp
#         Specifies the effective date of the schedule
#     end_dt : pandas.Timestamp
#         Specifies the termination date of the schedule
#     freq : str
#         Specify the period frequency
#     roll_conv : RollConv
#         How to treat dates that do not fall on a valid day. The default is RollConv.MODIFIED_FOLLOWING.
#     day_roll : DayRoll
#         Specifies the day periods should start/end on. The default is DayRoll.UNADJUSTED.
#     first_period_end: pandas.Timestamp
#         Specifies the end date of the first period. The first_period_end overrides the first_stub field.
#     last_period_start: pandas.Timestamp
#         Specifies the start date of the last period. The last_period_start overrides the last_stub field.
#     first_stub : Stub
#         Specifies the type of the first stub. If first_period_end is specified, the first_stub is ignored.
#     last_stub : Stub
#         Specifies the type of the last stub. If last_period_start is specified, the last_stub is ignored.
#     roll_user_specified_dates : bool
#         Boolean flag for whether to roll (per business day calendar and roll convention) the user specified dates (start_date, end_date, first_period_end, last_period_start)
#     cal : np.busdaycalendar
#         Specifies the business day calendar to observe.
#
#     Returns
#     -------
#     schedule : pandas.DataFrame
#         Columns:
#             - fixing_date (if add_fixing_dates=True)
#             - period_start
#             - period_end
#             - payment_date (if add_payment_dates=True)
#     """
#
#     start_date, end_date, first_period_end, last_period_start = [
#         pd.Timestamp(d) if isinstance(d, (np.datetime64, datetime.date, datetime.datetime)) else d
#         for d in [start_date, end_date, first_period_end, last_period_start]
#     ]
#
#     # Input validation
#     if start_date >= end_date:
#         raise ValueError(f"start_date {start_date} must be before end_date {end_date}")
#
#     if first_period_end is not None:
#         assert start_date < first_period_end <= end_date
#
#     if last_period_start is not None:
#         assert start_date <= last_period_start < end_date
#
#     # if freq == PeriodFreq.CDS:
#     #     pass
#     # elif freq == PeriodFreq.IMM:
#     #     pass
#
#     if freq == PeriodFreq.ZERO_COUPON or first_period_end == end_date or last_period_start == start_date:
#         # Function inputs explicitly specify 1 period
#         d1, d2 = [start_date], [end_date]
#     elif first_period_end == last_period_start and first_period_end is not None and last_period_start is not None:
#         # Function inputs explicitly specify 2 periods
#         d1 = [start_date, first_period_end]
#         d2 = [last_period_start, end_date]
#     else:
#         if first_period_end is None and last_period_start is None:
#             assert first_stub != Stub.DEFINED_PER_FIRST_PERIOD_END_DATE
#             assert last_stub != Stub.DEFINED_PER_LAST_PERIOD_START_DATE
#
#             if first_stub == Stub.DEFAULT and last_stub == Stub.DEFAULT:
#                 # If no stub is specified, defaults to market convention on the 1st stub, no last stub.
#                 first_stub = Stub.market_convention()
#                 last_stub = Stub.NONE
#             elif first_stub == Stub.DEFAULT:
#                 first_stub  = Stub.NONE
#             elif last_stub == Stub.DEFAULT:
#                 last_stub = Stub.NONE
#             else:
#                 raise ValueError("Unexpected logic branch - please raise GitHub issue")
#
#             if last_stub == Stub.NONE:
#                 d1, d2 = generate_date_schedule(start_date, end_date, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
#                 if first_stub == Stub.LONG:
#                     d1 = [d1[0]] + d1[2:]
#                     d2 = d2[1:]
#             elif first_stub == Stub.NONE and last_stub != Stub.NONE:
#                 d1, d2 = generate_date_schedule(start_date, end_date, freq, 'forward', roll_conv, day_roll, roll_user_specified_dates, cal)
#                 if last_stub == Stub.LONG:
#                     d1 = d1[:-1]
#                     d2 = d2[:-2] + [d2[-1]]
#             elif first_stub in [Stub.SHORT, Stub.LONG] and last_stub in [Stub.SHORT, Stub.LONG]:
#                 raise ValueError("If a schedule has first and last stubs they must be specified via first_period_end and last_period_start")
#             else:
#                 raise ValueError("Unexpected logic branch - please raise GitHub issue")
#
#         else:
#             # If first_period_end or last_period_start are specified we want to generate the schedule
#             # via generate_date_schedule(start_date, end_date, ...) and see if
#             # the first_period_end or last_period_start match the generated schedule.
#             #
#             # If they align, we use this generated schedule.
#             #
#             # If they don't align, we generate the schedule:
#             # (i) via backward generation from last_period_start if only last_period_start is specified
#             # (ii) via forward generation from first_period_end if only first_period_end is specified
#             # (iii) between first_period_end and last_period_start if both are specified
#
#             # Step 1: Generate the schedule by backward and forward date generation
#             d1_backward, d2_backward = generate_date_schedule(start_date, end_date, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
#             d1_forward, d2_forward = generate_date_schedule(start_date, end_date, freq, 'forward', roll_conv, day_roll, roll_user_specified_dates, cal)
#
#             # Step 2a : Check if the first_period_end matches any generated schedules
#
#             direction = []
#             if first_period_end is not None:
#                 if first_period_end == d2_forward[0]:
#                     direction.append('forward')
#                     first_stub = Stub.NONE
#                 elif first_period_end == d2_backward[0]:
#                     direction.append('backward')
#                     first_stub = Stub.SHORT
#                 elif first_period_end == d2_backward[1]:
#                     direction.append('backward')
#                     first_stub = Stub.LONG
#                 else:
#                     # Need to construct schedule using first_period_end
#                     first_stub = Stub.DEFINED_PER_FIRST_PERIOD_END_DATE
#
#             # Step 2a : Check if the last_period_start matches any generated schedules
#             if last_period_start is not None:
#                 if last_period_start == d1_backward[-1]:
#                     direction.append('backward')
#                     last_stub = Stub.NONE
#                 elif last_period_start == d1_forward[-1]:
#                     direction.append('forward')
#                     last_stub = Stub.SHORT
#                 elif last_period_start == d1_forward[-2]:
#                     direction.append('forward')
#                     last_stub = Stub.LONG
#                 else:
#                     # Need to construct schedule using last_period_start
#                     last_stub = Stub.DEFINED_PER_LAST_PERIOD_START_DATE
#
#             # Step 2c : Set default stub values to:
#             #              (i) NONE if other stub is SHORT/LONG
#             #              (ii) SHORT if other stub is NONE
#             if first_stub == Stub.DEFAULT and last_stub != Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
#                 assert last_stub != Stub.DEFAULT
#                 if last_stub == Stub.NONE:
#                     first_stub = Stub.market_convention()
#                 else:
#                     first_stub = Stub.NONE
#             if last_stub == Stub.DEFAULT and first_stub != Stub.DEFINED_PER_FIRST_PERIOD_END_DATE:
#                 assert first_stub != Stub.DEFAULT
#                 if first_stub == Stub.NONE:
#                     last_stub = Stub.market_convention()
#                 else:
#                     last_stub = Stub.NONE
#
#             # Step 3: Construct the schedules
#             if first_stub == Stub.DEFINED_PER_FIRST_PERIOD_END_DATE and last_stub == Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
#                 # Generate date schedule between first_period_end and last_period_start
#                 d1, d2 = generate_date_schedule(first_period_end, last_period_start, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
#                 d1 = [start_date] + d1 + [last_period_start]
#                 d2 = [first_period_end] + d2 + [end_date]
#             elif first_stub == Stub.DEFINED_PER_FIRST_PERIOD_END_DATE and last_stub != Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
#                 # Forward generation from first_period_end and if long last stub, combine last two periods
#                 d1, d2 = generate_date_schedule(first_period_end, end_date, freq, 'forward', roll_conv, day_roll, roll_user_specified_dates, cal)
#                 d1 = [start_date] + d1
#                 d2 = [first_period_end] + d2
#                 if last_stub == Stub.LONG:
#                     d1 = d1[:-1]
#                     d2 = d2[:-2] + [d2[-1]]
#             elif first_stub != Stub.DEFINED_PER_FIRST_PERIOD_END_DATE and last_stub == Stub.DEFINED_PER_LAST_PERIOD_START_DATE:
#                 # Backward generation from last_period_start and if long first stub, combine first two periods
#                 d1, d2 = generate_date_schedule(start_date, last_period_start, freq, 'backward', roll_conv, day_roll, roll_user_specified_dates, cal)
#                 d1 = d1 + [last_period_start]
#                 d2 = d2 + [end_date]
#                 if first_stub == Stub.LONG:
#                     d1 = [d1[0]] + d1[2:]
#                     d2 = d2[1:]
#             else:
#                 # Don't need to generate any additional schedules
#                 assert len(set(direction)) == 1
#
#                 if first_stub == Stub.NONE:
#                     # Forward date generation
#                     d1, d2 = d1_forward, d2_forward
#                     if last_stub == Stub.LONG:
#                         d1 = d1[:-1]
#                         d2 = d2[:-2] + [d2[-1]]
#                 elif last_stub == Stub.NONE:
#                     # Backward date generation
#                     d1, d2 = d1_backward, d2_backward
#                     if first_stub == Stub.LONG:
#                         d1 = [d1[0]] + d1[2:]
#                         d2 = d2[1:]
#                 else:
#                     raise ValueError("Unexpected logic branch - please raise GitHub issue")
#
#     schedule = pd.DataFrame({'period_start': d1, 'period_end': d2})
#     return schedule
#
#
# def generate_date_schedule(
#         start_date: pd.Timestamp,
#         end_date: pd.Timestamp,
#         freq: PeriodFreq,
#         direction: str,
#         roll_conv: RollConv=RollConv.MODIFIED_FOLLOWING,
#         day_roll: DayRoll=DayRoll.UNADJUSTED,
#         roll_user_specified_dates: bool=False,
#         cal: np.busdaycalendar=np.busdaycalendar()
#     ) -> Tuple[List, List]:
#     """
#     Generates a schedule of start and end dates between start_date and end_date.
#
#     Parameters
#     ----------
#     start_date : pd.Timestamp
#         The start date of the schedule.
#     end_date : pd.Timestamp
#         The end date of the schedule.
#     freq : PeriodFreq
#         The frequency of the schedule.
#     direction : {'forward', 'backward'}
#         The direction in which to generate dates.
#     roll_conv : RollConv
#         How to treat dates that fall on a non business day.
#     day_roll : DayRoll
#         Specifies the day periods should start/end on.
#     roll_user_specified_dates : bool
#         Boolean flag for whether to roll (per business day calendar and roll convention) the user specified dates (start_date, end_date)
#     cal : np.busdaycalendar
#         Specifies the business day calendar to observe.
#
#     Returns
#     -------
#     Tuple[List, List]
#
#     Raises
#     ------
#     ValueError, TypeError
#         If any of the inputs have invalid values or types
#
#     TODO: Consider options for passing param=(start_date, end_date, freq, None, None, None, None, None, None, None, None) to function
#           E.g. Wrap with fields with Optional[] and set defaults in function from enum.set_default()
#     """
#
#     def busday_offset_timestamp(pd_timestamp, offsets, roll_conv, cal):
#         np_datetime64D = np.array([pd_timestamp]).astype('datetime64[D]')
#         rolled_date_np = np.busday_offset(np_datetime64D, offsets=offsets, roll=roll_conv, busdaycal=cal)[0]
#         return pd.Timestamp(rolled_date_np)
#
#     def apply_specific_day_roll(pd_timestamp: pd.Timestamp,
#                                 specific_day_roll: DayRoll) -> pd.Timestamp:
#         if specific_day_roll == DayRoll.UNADJUSTED:
#             return pd_timestamp
#         else:
#             try:
#                 return pd_timestamp.replace(day=specific_day_roll.value)
#             except ValueError:
#                 return pd_timestamp + pd.offsets.MonthEnd(0)
#
#     # Input validation
#     if direction not in {'forward', 'backward'}:
#         raise ValueError(f"Invalid direction '{direction}'. Must be 'forward' or 'backward'.")
#     if not isinstance(start_date, pd.Timestamp) or not isinstance(end_date, pd.Timestamp):
#         raise TypeError("'start_date' {start_date} and 'end_date' {end_date} must be pandas Timestamp objects.")
#     if start_date >= end_date:
#         raise ValueError("'start_date' must be earlier than 'end_date'.")
#
#     # Implementation rationale:
#     # - append to both start_dates and end_dates in while loop so don't have to consider edge case of list length = 1
#     # - Note the 'roll day' component must be applied inside the while loop due to holidays/weekends
#     #   which could move a day past the while loop condition.
#     i = 1
#     start_dates = []
#     end_dates = []
#     if roll_user_specified_dates:
#         start_date_in_schedule = busday_offset_timestamp(start_date, 0, roll_conv.value, cal)
#         end_date_in_schedule = busday_offset_timestamp(end_date, 0, roll_conv.value, cal)
#     else:
#         start_date_in_schedule = start_date
#         end_date_in_schedule = end_date
#
#     if direction == 'forward':
#         current_date = apply_specific_day_roll(start_date + freq.date_offset, day_roll)
#         current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
#
#         while current_date < end_date:
#             start_dates.append(current_date)
#             end_dates.append(current_date)
#             current_date = start_date + freq.multiply_date_offset(i+1)
#             current_date = apply_specific_day_roll(current_date, day_roll)
#             current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
#             i += 1
#
#         start_dates = [start_date_in_schedule] + start_dates
#         end_dates.append(end_date_in_schedule)
#
#     elif direction == 'backward':
#         current_date = apply_specific_day_roll(end_date - freq.date_offset, day_roll)
#         current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
#
#         while current_date > start_date:
#             start_dates.append(current_date)
#             end_dates.append(current_date)
#             current_date = end_date - freq.multiply_date_offset(i+1)
#             current_date = apply_specific_day_roll(current_date, day_roll)
#             current_date = busday_offset_timestamp(current_date, 0, roll_conv.value, cal)
#             i +=1
#
#         start_dates.append(start_date_in_schedule)
#         end_dates = [end_date_in_schedule] + end_dates
#
#         start_dates.reverse()
#         end_dates.reverse()
#
#     return start_dates, end_dates
#
#
# def create_date_grid_for_fx_exposures(curve_date: pd.Timestamp,
#                                       delivery_dates: np.array,
#                                       sampling_freq: str=None,
#                                       date_grid_pillar: pd.DatetimeIndex=None,
#                                       payments_on_value_date_have_value: bool=False,
#                                       include_last_day_of_value: bool=False,
#                                       include_day_after_last_day_of_value: bool=False) -> pd.DatetimeIndex:
#     """
#     Defines a date grid for calculating exposures
#     Please note, for each trade, this defines the settlement date grid, hence the expiry
#     date grid needs to be solved based on the delay of the market.
#     For FX derivatives this is relively simple as market data inputs can be interploted from
#     (i) expiry or (ii) settle dates.
#
#     Parameters
#     ----------
#     curve_date : pd.Timestamp
#         DESCRIPTION.
#     delivery_dates : np.array
#         DESCRIPTION.
#     sampling_freq : str, optional
#         DESCRIPTION. The default is None.
#     date_grid_pillar : pd.DatetimeIndex, optional
#         DESCRIPTION. The default is None.
#     include_payments_on_value_date : bool, optional
#         DESCRIPTION. The default is False.
#     include_last_day_of_value : bool, optional
#         DESCRIPTION. The default is False.
#     include_day_after_last_day_of_value : bool, optional
#         DESCRIPTION. The default is False.
#
#     Returns
#     -------
#     date_grid : TYPE
#         DESCRIPTION.
#
#     """
#
#     # Challenge is that exposures are defined using the rate-fixing - i.e the expiry date
#     # The expiry dates are what we need to specify for our FX option valuation or FX rate simulation
#     # The trade technically still has value until some time on day of delivery/settlement.
#     # Hence the date grid should be constructed on a delivery/settlement basis.
#     # There is a pototo/potata view on if a cashflow has value on the settlement date.
#     # Our default view is no, it should be in accounts receivable, but this can be toggled.
#
#     delivery_dates = np.unique(delivery_dates)
#     max_settlement_date = delivery_dates.max()
#
#     if sampling_freq is not None:
#         if sampling_freq == '1d':
#             date_grid = generate_date_schedule(curve_date, max_settlement_date, 1, 'days', 'forward')
#         elif sampling_freq == '1w':
#             date_grid = generate_date_schedule(curve_date, max_settlement_date, 1, 'weeks', 'forward')
#         elif sampling_freq == '1m':
#             date_grid = generate_date_schedule(curve_date, max_settlement_date, 1, 'months', 'forward')
#         elif sampling_freq == '3m':
#             date_grid = generate_date_schedule(curve_date, max_settlement_date, 3, 'months', 'forward')
#         elif sampling_freq == '6m':
#              date_grid = generate_date_schedule(curve_date, max_settlement_date, 6, 'months', 'forward')
#         elif sampling_freq == '12m':
#              date_grid = generate_date_schedule(curve_date, max_settlement_date, 12, 'months', 'forward')
#
#         date_grid = pd.DatetimeIndex([curve_date]).append(pd.DatetimeIndex(date_grid[1]))
#
#     elif date_grid_pillar is not None:
#         assert isinstance(date_grid_pillar, pd.DatetimeIndex)
#         date_grid = date_grid_pillar
#     else:
#         raise ValueError("Either 'sampling_freq' or date_grid_pillar' must be specified")
#
#     if include_last_day_of_value:
#         if payments_on_value_date_have_value:
#             last_day_of_value = delivery_dates
#         else:
#             last_day_of_value = delivery_dates - pd.DateOffset(days=1)
#         date_grid = pd.DatetimeIndex(np.concatenate(date_grid.values, last_day_of_value))
#
#     elif include_day_after_last_day_of_value:
#         if payments_on_value_date_have_value:
#             last_day_of_value = delivery_dates + pd.DateOffset(days=1)
#         else:
#             last_day_of_value = delivery_dates
#         date_grid = pd.DatetimeIndex(np.concatenate(date_grid.values, last_day_of_value))
#
#     date_grid = date_grid.drop_duplicates().sort_values()
#     return date_grid
#



# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

import numpy as np
import time

from frm.pricing_engine.monte_carlo_generic import generate_rand_nbs


def simulate_gbm_path(initial_px: np.array,
                      drift: np.array,
                      forward_volatility: np.array,
                      timestep_length: np.array,
                      rand_nbs: np.array):
    """
    Simulates the path of a Geometric Brownian Motion (GBM) over a defined period set.

    This function simulates the path of a stochastic process following GBM, given initial prices,
    drift rates, volatilities, and random numbers for each period. It supports vectorized operations
    allowing simulation of multiple paths in parallel.

    Parameters:
    initial_px (np.array): An array of initial prices for each simulation path.
    drift (np.array): An array of drift rates for each time step.
    forward_volatility (np.array): An array of volatilities for each time step.
    timestep_length (np.array): An array representing the length of each step
    rand_nbs (np.array): A 3D array of random numbers; shape=(number of time steps, number of random variables, number of simulations)

    Returns:
    np.array: A 2D array where each row represents a simulated GBM path and each column
              represents a time step in the path.

    Raises:
    AssertionError: If the shapes of period_length, period_drift, period_forward_volatility,
                    and rand_nbs do not align as expected.

    Notes:
    - The function assumes that period_length, period_drift, and period_forward_volatility have
      the same shape and that the number of periods matches the number of columns in rand_nbs.
    - It simulates the path by sequentially updating the price for each period using the GBM formula.
    """

    assert timestep_length.ndim == 1

    # Reshape inputs to 2D arrays so to support parallel simulation of multiple underlyings
    drift = np.atleast_2d(drift)
    forward_volatility = np.atleast_2d(forward_volatility)
    initial_px = np.atleast_2d(initial_px)

    # if drift.ndim == 1:
    #     drift = drift.reshape(-1, 1)
    # if forward_volatility.ndim == 1:
    #     forward_volatility = forward_volatility.reshape(-1, 1)
    # if initial_px.ndim == 1:
    #     initial_px = initial_px.reshape(1, -1)

    if drift.shape[0] == 1:
        drift = drift.reshape(-1, 1)
    if forward_volatility.shape[0] == 1:
        forward_volatility = forward_volatility.reshape(-1, 1)
    if initial_px.shape[0] == 1:
        initial_px = initial_px.reshape(1, -1)

    # Shape checks
    nb_timesteps = timestep_length.shape[0]
    assert nb_timesteps == drift.shape[0]
    assert nb_timesteps == forward_volatility.shape[0]
    assert nb_timesteps == rand_nbs.shape[0]
    assert initial_px.shape[0] == 1

    nb_rand_vars = initial_px.shape[1]
    assert nb_rand_vars == drift.shape[1]
    assert nb_rand_vars == forward_volatility.shape[1]
    assert nb_rand_vars == rand_nbs.shape[1]

    # Initialise shape and set initial value of underlying(s)
    x = np.zeros((rand_nbs.shape[0]+1, rand_nbs.shape[1], rand_nbs.shape[2]))
    x[0, :, :] = initial_px

    # Simulate per geometric brownian motion process
    for i in range(0, nb_timesteps):
        x[i+1, :, :] = x[i, :, :] * np.exp((drift[i, :] - 0.5 * (forward_volatility[i, :] ** 2)) *
                                           timestep_length[i] + forward_volatility[i, :] * np.sqrt(timestep_length[i]) * rand_nbs[i, :, :])

    return x


if __name__ == "__main__":

    initial_px = np.array([0.6629])
    drift = np.array([0.95*0.01, 0.55*0.01, -0.02*0.01, -0.34*0.01])
    forward_volatility = np.array([9.73*0.01, 10.10*0.01, 9.75*0.01, 10.97*0.01])
    timestep_length = np.array([0.512328767, 0.498630137, 0.501369863, 0.498630137])
    nb_steps = len(timestep_length)

    rand_nbs = generate_rand_nbs(nb_steps=nb_steps,
                                 nb_rand_vars=1,
                                 nb_simulations=1000 * 1000)


    t1 = time.time()
    results = simulate_gbm_path(initial_px=initial_px,
                                drift=drift,
                                forward_volatility=forward_volatility,
                                timestep_length=timestep_length,
                                rand_nbs=rand_nbs)

    t2 = time.time()
    
    print("GBM:", t2-t1)




# %%






# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

import re 
import unicodedata
import logging
import numpy as np
import pandas as pd
from pandas import DateOffset
from typing import Optional


def clean_tenor(tenor: str) -> str:
    if not isinstance(tenor, str):
        raise TypeError(f"'tenor' {tenor} must be a string. Instead is type {type(tenor)}")
    
    tenor = unicodedata.normalize('NFKD', tenor)
    tenor = tenor.lower().replace(' ','').replace('/','').replace('\n', '').replace('\r', '')

    replacements = {
        'd': ['days', 'day'],
        'w': ['weeks', 'week'],
        'm': ['months', 'month', 'mon'],
        'y': ['years', 'year', 'yrs', 'yr'],
        'on': ['overnight'],
        'tn': ['tomorrownext', 'tomnext'],
        'sw': ['spotweek'],
        'sn': ['spotnext'],
        'sp': ['spot']
    }

    pattern = re.compile('|'.join(map(re.escape, [val for sublist in replacements.values() for val in sublist])))
    tenor = pattern.sub(lambda match: next(k for k, v in replacements.items() if match.group(0) in v), tenor)

    return tenor


def tenor_to_date_offset(tenor: str) -> pd.DateOffset:    
    if not isinstance(tenor, str):
        raise TypeError("'tenor' must be a string")      
        
    misc_tenors_offset = {
        'on' : DateOffset(days=1), # 1 day (overnight)
        'tn' : DateOffset(days=2), # 2 days (tomorrow next)
        'sp'  : DateOffset(days=0), # spot date
        'sn' : DateOffset(days=1), # 1 day (spot next)
        'sw' : DateOffset(days=5), # 5 days (spot week)        
        }
    
    if tenor in misc_tenors_offset.keys():
        offset = misc_tenors_offset[tenor]
    else:
        # Identity tenors specified in integer days only; 1D, 30D , 360D
        if re.search('^\d+d$',tenor) is not None:
            offset = DateOffset(days=int(tenor[:-1]))
        # Identity tenors specified in integer weeks only; 1W, 52W, 104W
        elif re.search('^\d+w$',tenor) is not None:
            offset = DateOffset(weeks=int(tenor[:-1]))
        # Identity tenors specified in integer months only; 1M, 12M, 120M
        elif re.search('^\d+m$',tenor) is not None:
            offset = DateOffset(months=int(tenor[:-1]))       
        # Identity tenors specified in integer years only; 1Y, 10Y, 100Y
        elif re.search('^\d+y$',tenor) is not None:
            offset = DateOffset(years=int(tenor[:-1]))   
        # Identity tenors specified in integer years and integer monthly only; 1Y3M, 10Y6M, 100Y1M
        elif re.search('^\d+y\d+m$',tenor) is not None:
            years,months = tenor[:-1].split('y')
            total_months = int(years)*12 + int(months)
            offset = DateOffset(months=total_months)   
        else: 
            logging.error(f"invalid 'tenor' value: {tenor}")      
            raise ValueError(f"invalid 'tenor' value: {tenor}" )
        
    return offset

#
# np.busday_offset(
#     dates=(df['effective_date'] + tenor_date_offset).to_numpy().astype('datetime64[D]'), offsets=0,
#     roll='following', busdaycal=busdaycal))

def workday(dates: [pd.Timestamp, pd.DatetimeIndex],
            offset: int,
            cal: np.busdaycalendar) -> pd.Timestamp:

    # Convert to datetime64[D] to use in np.busday_offset
    dates = dates.to_numpy().astype('datetime64[D]')
    return np.busday_offset(dates=dates, offsets=offset, roll='following', busdaycal=cal)




def check_day_offset_consistency(
        date: pd.Timestamp,
        offset_date: pd.Timestamp,
        day_offset: int,
        busdaycal: np.busdaycalendar):

    implied_offset = calc_implied_business_day_offset(date, offset_date, busdaycal)
    if implied_offset != day_offset:
        raise ValueError(f"The implied offset is {implied_offset} "
                         f"which is inconsistent with the provided day offset of {day_offset}.")


def calc_implied_business_day_offset(
        date: pd.Timestamp,
        offset_date: pd.Timestamp,
        cal: np.busdaycalendar) -> int:
    assert date >= offset_date, "date must be greater than or equal to offset_date"

    offset = 0
    while date < offset_date:
        offset += 1
        date += pd.DateOffset(days=1)
        date_np = date.to_numpy().astype('datetime64[D]')
        date = np.busday_offset(date_np, offsets=0, roll='following', busdaycal=cal)
    return offset


def resolve_fx_curve_dates(
        ccy_pair: str,
        cal: np.busdaycalendar,
        curve_date: Optional[pd.Timestamp] = None,
        spot_offset: Optional[int] = None,
        spot_date: Optional[pd.Timestamp] = None):

    if curve_date is not None and spot_date is not None and spot_offset is not None:
        # All three dates are specified, check for consistency
        check_day_offset_consistency(date=curve_date, offset_date=spot_date, offset=spot_offset, busdaycal=cal)
    elif spot_offset is None and curve_date is not None and spot_date is not None:
        # Spot offset is not specified, calculate it based on curve date and spot date
        spot_offset = calc_implied_business_day_offset(curve_date, spot_date, cal)
    else:
        # Only one of {curve_date, spot_date} are specified, get spot_offset per market convention from ccy_pair
        spot_offset = get_fx_spot_day_offset(ccy_pair)

    if spot_date is None:
        curve_date_np = curve_date.to_numpy().astype('datetime64[D]')
        spot_date = pd.Timestamp(
            np.busday_offset(curve_date_np, offsets=spot_offset, roll='following', busdaycal=cal))
    elif curve_date is None:
        spot_date_np = spot_date.to_numpy().astype('datetime64[D]')
        curve_date = pd.Timestamp(
            np.busday_offset(spot_date_np, offsets=-spot_offset, roll='preceding', busdaycal=cal))

    return curve_date, spot_offset, spot_date


def get_fx_spot_day_offset(validated_ccy_pair: str) -> int:

    if len(validated_ccy_pair) == 6:
        # http://www.londonfx.co.uk/valdates.html
        # https://web.archive.org/web/20240213223033/http://www.londonfx.co.uk/valdates.html

        if validated_ccy_pair in {'usdcad','cadusd',
                                  'usdphp','phpusd',
                                  'usdrub','rubusd',
                                  'usdtry','tryusd'}:
            return 1
        else:
            return 2


def get_ir_settlement_offset(ccy: str) -> int:
    if ccy in {'usd', 'cad'}:
        return 2
    elif ccy in {'aud', 'nzd', 'jpy', 'eur'}:
        return 2
    elif ccy in {'gbp'}:
        return 0
    else:
        return 2


# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
import re
from typing import Union
import warnings
from frm.utils.utilities import convert_column_to_consistent_data_type
from frm.utils.tenor import clean_tenor, tenor_to_date_offset

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM')) 

VALID_DELTA_CONVENTIONS = ['regular_spot','regular_forward','premium_adjusted_spot','premium_adjusted_forward']



def fx_term_structure_helper(df: pd.DataFrame,
                             curve_date: pd.Timestamp,
                             spot_offset: int,
                             cal: np.busdaycalendar,
                             rate_set_date_str: str) -> pd.DataFrame:

    df['tenor'] = df['tenor'].apply(clean_tenor)

    mandatory_columns_at_least_one_of = [['tenor',rate_set_date_str,'delivery_date']]
    for column_set in mandatory_columns_at_least_one_of:
        assert sum([column in df.columns for column in column_set]) >= 1, f'Specify at least one of {column_set}'

    for column in [rate_set_date_str,'delivery_date']:
        if column not in df.columns:
            df[column] = np.nan
        df[column] = df[column].astype('datetime64[ns]')

    # Input hierarchy is:
    # 1. Use the delivery date if it is available.
    # 2. Otherwise, calculate the delivery date from the rate-set date ('fixing_date' for forwards, 'expiry_date' for options
    # 3. Otherwise, calculate the rate-set date and delivery date from the tenor
    for i, row in df.iterrows():
        if pd.notna(row['delivery_date']):
            pass
        elif pd.notna(row[rate_set_date_str]):
            rate_set_date_np = row[rate_set_date_str].astype('datetime64[D]')
            df.at[i,'delivery_date'] = np.busday_offset(rate_set_date_np, offsets=-1*spot_offset, roll='preceding',busdaycal=cal)
        elif pd.notna(row['tenor']):
            date_offset = tenor_to_date_offset(row['tenor'])
            rate_set_date = curve_date + date_offset
            rate_set_date_np = rate_set_date.to_numpy().astype('datetime64[D]')
            df.at[i,rate_set_date_str] = np.busday_offset(rate_set_date_np, offsets=0, roll='following',busdaycal=cal)
            df.at[i,'delivery_date'] = np.busday_offset(rate_set_date_np, offsets=spot_offset, roll='following',busdaycal=cal)
        else:
            raise ValueError("'delivery_date', '{rate_set_date_str}' and 'tenor' are all missing")

    if 'tenor' in df.columns:
        time_column_order = ['tenor'] + [rate_set_date_str,'delivery_date']
    else:
        time_column_order = [rate_set_date_str,'delivery_date']
    column_order = time_column_order + [col for col in df.columns if col not in time_column_order]

    df = df[column_order]
    df = df.sort_values('delivery_date').reset_index(drop=True)

    return df


def fx_forward_curve_helper(fx_forward_curve_df: pd.DataFrame,
                            curve_date: pd.Timestamp,
                            spot_offset: int,
                            cal: np.busdaycalendar) -> pd.DataFrame:

    fx_forward_curve_df = convert_column_to_consistent_data_type(fx_forward_curve_df)

    mandatory_columns = ['fx_forward_rate']
    for column in mandatory_columns:
        assert column in fx_forward_curve_df.columns, f'{column} is missing in fx_forward_curve'

    fx_forward_curve_df = fx_term_structure_helper(df=fx_forward_curve_df,
                                                   curve_date=curve_date,
                                                   spot_offset=spot_offset,
                                                   cal=cal,
                                                   rate_set_date_str='fixing_date')
    return fx_forward_curve_df
    
    
def validate_ccy_pair(ccy_pair):

    ccy_pair = ccy_pair.lower().replace('/','').strip()
    assert len(ccy_pair) == 6, ccy_pair
    foreign_ccy = ccy_pair[:3] # Also known as the base currency
    domestic_ccy = ccy_pair[3:] # Also known as the quote currency

    if foreign_ccy == 'usd' and domestic_ccy in {'aud', 'eur', 'gbp', 'nzd'}:
        warnings.warn(
            "non conventional fx market, market convention is 'usd' as the domestic currency, i.e. the the quote currency, for 'aud/usd', 'eur/usd', 'gbp/usd' and 'nzd/usd' pairs")
    elif domestic_ccy == 'usd' and foreign_ccy not in {'aud', 'eur', 'gbp', 'nzd'}:
        warnings.warn(
            "non conventional fx market, market convention is for 'usd' to be the foreign currency, i.e. the base currency, (except for 'aud/usd', 'eur/usd', 'gbp/usd' and 'nzd/usd' pairs)")

    return ccy_pair, domestic_ccy, foreign_ccy





def check_delta_convention(df: pd.DataFrame, ccy_pair: str) -> pd.DataFrame:

    bool_cond = np.logical_and(ccy_pair in ['audusd', 'nzdeur', 'gbpusd', 'nzdusd'],
                               df['delta_convention'].str.contains('premium_adjusted'))
    df.loc[bool_cond, 'warnings'] += 'the regular delta is the market delta_convention for this currency pair\n'
    if bool_cond.any():
        warnings.warn('The delta_convention is premium_adjusted for the following tenors: '
                      f'{df.loc[bool_cond, "tenor"].unique()} for the currency pair {ccy_pair}. '
                      'The market convention is to use the regular delta. '
                      'Please ensure that the delta_convention is consistent with the market convention.')

    conditions_for_premium_adjusted_delta_convention = [
        np.full(len(df), np.logical_not(ccy_pair in ['audusd', 'nzdeur', 'gbpusd', 'nzdusd', 'usdaud', 'usdeur', 'usdgbp', 'usdnzd'])),
        np.full(len(df), 'usd' in ccy_pair),
        df['delta_convention'].str.contains('regular').values]
    bool_cond = np.logical_and.reduce(conditions_for_premium_adjusted_delta_convention)
    df.loc[bool_cond, 'warnings'] += 'premium adjusted delta is the market delta_convention for this currency pair\n'

    if bool_cond.any():
        warnings.warn('The delta_convention is regular for the following tenors: '
                      f'{df.loc[bool_cond, "tenor"].unique()} for the currency pair {ccy_pair}. '
                      'The market convention is to use the premium adjusted delta. '
                      'Please ensure that the delta_convention is consistent with the market convention.')


    return df


def clean_vol_quotes_column_names(vol_quotes_df):
    patterns = {
        r'^(0?[1-9]|[1-4]\d)[_ ]?(delta|Δ)[_ ]?(call|c|put|p)$': lambda
            m: f'{m.group(1)}_delta_{"call" if m.group(3) in ["call", "c"] else "put"}',
        r'^(0?[1-9]|[1-4]\d)[_ ]?(delta|Δ)[_ ]?(risk[_ ]?reversal|rr|butterfly|bf)$': lambda
            m: f'{m.group(1)}_delta_{m.group(3).replace("_", "").replace(" ", "").replace("rr", "riskreversal").replace("bf", "butterfly")}',
        r'^(a|at)[_ ]?(t|the)[_ ]?(m|money)[_ ]?(delta|Δ)[_ ]?neutral$': lambda _: 'atm_delta_neutral',
        r'^(a|at)[_ ]?(t|the)[_ ]?(m|money)[_ ]?(f|fwd|forward)$': lambda _: 'atm_forward'
    }

    for col_name in vol_quotes_df.columns:
        for pattern, func in patterns.items():
            match = re.match(pattern, col_name)
            if match:
                new_col_name = func(match)
                vol_quotes_df.rename(columns={col_name: new_col_name}, inplace=True)
                break

    return vol_quotes_df


def solve_call_put_quotes_from_strategy_quotes(vol_quotes_df):

    risk_reversal_deltas = sorted([col_name.split('_')[0] for col_name in vol_quotes_df.columns if 'riskreversal' in col_name])
    butterfly_deltas = sorted([col_name.split('_')[0] for col_name in vol_quotes_df.columns if 'butterfly' in col_name])

    assert 'atm_delta_neutral' in vol_quotes_df.columns
    assert risk_reversal_deltas == butterfly_deltas

    for delta in butterfly_deltas:

        if f'{delta}_delta_put' not in vol_quotes_df.columns:
            vol_quotes_df[f'{delta}_delta_put'] = vol_quotes_df['atm_delta_neutral'] \
                                                        + vol_quotes_df[f'{delta}_delta_butterfly'].values \
                                                        - 0.5 * vol_quotes_df[f'{delta}_delta_riskreversal'].values
        if f'{delta}_delta_call' not in vol_quotes_df.columns:
            vol_quotes_df[f'{delta}_delta_call'] = vol_quotes_df['atm_delta_neutral'] \
                                                         + vol_quotes_df[f'{delta}_delta_butterfly'].values \
                                                         + 0.5 * vol_quotes_df[f'{delta}_delta_riskreversal'].values

        vol_quotes_df.drop(labels=[f'{delta}_delta_butterfly', f'{delta}_delta_riskreversal'], axis=1, inplace=True)

    return vol_quotes_df


def get_delta_smile_quote_details(df: pd.DataFrame,
                                  call_put_pattern: str = r'^(0?[1-9]|[1-4]\d)[_ ]?(delta|Δ)[_ ]?(call|c|put|p)$',
                                  atm_delta_neutral_column_pattern: str = r'^(a|at)[_ ]?(t|the)[_ ]?(m|money)[_ ]?(delta|Δ)[_ ]?neutral$') -> pd.DataFrame:
    quotes_column_names = []
    for col_name in df.columns:
        if re.match(call_put_pattern, col_name):
            quotes_column_names.append(col_name)
        elif re.match(atm_delta_neutral_column_pattern, col_name):
            quotes_column_names.append(col_name)

    call_put_flag = np.array([1 if 'call' in col_name else -1 if 'put' in col_name else 1 if col_name == 'atm_delta_neutral' else np.nan for col_name in quotes_column_names])

    signed_delta = np.array(
        [0.5 if col_name == 'atm_delta_neutral' else call_put_flag[i] * float(col_name.split('_')[0]) / 100 for i, col_name
         in enumerate(quotes_column_names)]
    )
    delta_smile_quote_details = pd.DataFrame({'quotes_column_names': quotes_column_names, 'quotes_call_put_flag': call_put_flag, 'quotes_signed_delta': signed_delta})
    delta_smile_quote_details['order'] = 100 / delta_smile_quote_details['quotes_signed_delta']
    delta_smile_quote_details = delta_smile_quote_details.sort_values('order', ascending=True).reset_index(drop=True)
    delta_smile_quote_details.drop(labels='order', axis=1, inplace=True)

    return delta_smile_quote_details


def interp_fx_forward_curve_df(
        fx_forward_curve_df: pd.DataFrame,
        dates: Union[pd.Series, pd.DatetimeIndex] = None,
        date_type: str = None,
        flat_extrapolation: bool = True) -> pd.Series:

    assert date_type in ['fixing_date', 'delivery_date']

    if isinstance(dates, pd.Series):
        dates = pd.DatetimeIndex(dates)
    elif isinstance(dates, pd.DatetimeIndex):
        pass
    else:
        raise ValueError("'dates' must be a pandas Series or DatetimeIndex")

    unique_dates = pd.DatetimeIndex(dates.drop_duplicates())
    combined_dates = unique_dates.union(pd.DatetimeIndex(fx_forward_curve_df[date_type]))
    df = pd.Series(fx_forward_curve_df['fx_forward_rate'].values, index=fx_forward_curve_df[date_type].values)

    result = df.reindex(combined_dates.values).copy()
    start_date = fx_forward_curve_df[date_type].min()
    end_date = fx_forward_curve_df[date_type].max()

    if flat_extrapolation:
        try:
            result = result.interpolate(method='time', limit_area='inside').ffill().bfill()
        except ValueError:
            raise ValueError("Interpolation failed, check for missing values in the input data.")
        # Find out of range dates and warn
        out_of_range_dates = unique_dates[(unique_dates < start_date) | (unique_dates > end_date)]
        for date in out_of_range_dates:
            warnings.warn(
                f"Date {date} is outside the {date_type} range {start_date} - {end_date}, flat extrapolation applied.")
    else:
        result = result.interpolate(method='time', limit_area='inside')

    return result.reindex(dates).values





# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, InitVar
from enum import Enum
import numpy as np
import os
import pandas as pd
from typing import Optional, Union
import warnings

import frm.utils
from frm.utils import Schedule, get_schedule, day_count, year_frac, get_busdaycal
from frm.enums import CompoundingFreq, TermRate, RFRFixingCalcMethod, PeriodFreq, DayCountBasis, ExchangeNotionals
from frm.term_structures.zero_curve import ZeroCurve
from frm.term_structures.zero_curve_helpers import discount_factor_from_zero_rate
from scipy.optimize import root_scalar


from frm.instruments.leg import FixedLeg, FloatTermLeg, FloatRFRLeg, ZerocouponLeg

if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))

# Swap class
# Two legs, pay leg and rcv leg.

# Multi curve class
# It is possible to use zero curves + fixings though there will be lots of fiddly bits to get right:
# - DV01 (if swap uses multiple curves of same currency)
# - the kinked AUD swap curve (for swaption term structure construction / IRS bootstrapping
# - converting a swaption vol surface from 3M to 6M using basis spreads.
# - keeping track of basis spread curves.
# - stub coupon calculation - use part of the 1M and 3M curves.
# Hence, in a multi-curve class we would:
# - have helpers for G10 swap curves
# - do global DV01 shift across AONIA, BBSW 1M/3M/6M.
# - bootstrap
# - +/- 100 bps shift to par rates or DV01 done on par rates
# We want to be able to add various rates:

# AONIA
# (a) AONIA swap rates
# (b) AONIA and BBSW 3M basis spreads

# Defining the 3M and 6M curves:
# (a) BBSW 3M swap rates for <= 3Y and BBSW 6M swap rates for > 3Y and basis spreads
# (b) BBSW 6M swap rates and 3M/6M basis spreads
# (c) BBSW 3M swap rates and 3M/6M basis spreads
# (d) BBSW 6M swap rates and BBSW 3M swap rates

# Defining the 1M curve.
# BBSW 1m/3m basis spreads

# Defining the FCB curve
# AONIA/BBSW 3M basis spread curve
# AONIA vs SOFR FCB curve, BBSW 3M vs SOFR FCB curve

# The user defines buckets of quotes:
# The quotes either (i) purely define a zero curve or (ii) define the basis between two curves.
# It is probably best to "hard code" the process.

# Method:
# Choice (a) OIS DC Stripping? Requires to have fixed rates for OIS curve if doing a iterative bootstrap solve.
# For each instrument defining a solve user needs to define
# (i) if the curve is a forward curve only solve
# (ii) or fwd and discount curve solve (i.e. AONIA, or BBSW 3M under pure IBOR curve)
# (iii) or discount curve only solve (for FCB)
# If it is a forward curve solve, the user needs to the discount curve.



@dataclass
class Swap(ABC):
    "Generic class to cover pricing of all types of swap legs and bonds"

    # Required parameters
    leg1: Union[FixedLeg, FloatTermLeg, FloatRFRLeg, ZerocouponLeg]
    leg2: Union[FixedLeg, FloatTermLeg, FloatRFRLeg, ZerocouponLeg]

    # Optional parameters. Potentially move into an "Instrument" class.
    trade_date: Optional[pd.Timestamp] = None
    trade_price: Optional[float] = 0
    trade_id: Optional[str] = None
    trade_comment: Optional[str] = None
    # own_cds_curve: Optional[CDSCurve] = None
    # counterparty_cds_curve: Optional[CDSCurve] = None

    def calc_dirty_pv(self):
        leg1_pv = self.leg1.calc_dirty_pv()
        leg2_pv = self.leg2.calc_dirty_pv()
        return leg1_pv + leg2_pv

    def calc_accrued_interest(self):
        leg1_accrued_interest = self.leg1.calc_accrued_interest()
        leg2_accrued_interest = self.leg2.calc_accrued_interest()
        return leg1_accrued_interest + leg2_accrued_interest

    def calc_clean_pv(self):
        return self.calc_dirty_pv() - self.calc_accrued_interest()

    def par_solve_each_leg(self):
        leg1_par = self.leg1.par_solve()
        leg2_par = self.leg2.par_solve()
        return leg1_par, leg2_par

    def target_solve_swap(self, target_pv: float, leg_number: int):
        if leg_number == 1:
            return self.leg1.target_solve(target_pv)
        elif leg_number == 2:
            return self.leg2.target_solve(target_pv)
        else:
            raise ValueError('leg_number must be 1 or 2')

    def __post_init__(self):
        assert self.leg1.pay_rcv != self.leg2.pay_rcv, 'Legs must have opposite pay/receive perspectives'


# Potentially, create bunch of classes for different types of swaps that fill in default values (i.e ZC swap, Fixed/Float swap, etc)
# Then, for common quote types, can define all the "non-quote" parameters in the class and then pass in the quote parameters.


# -*- coding: utf-8 -*-
import os
if __name__ == "__main__":
    os.chdir(os.environ.get('PROJECT_DIR_FRM'))
    
from enum import Enum


class RFRFixingCalcMethod(Enum):
    DAILY_COMPOUNDED = 'daily_compounded'
    WEIGHTED_AVERAGE = 'weighted_average'
    SIMPLE_AVERAGE = 'simple_average'

    @property
    def display_name(self):
        return self.name.replace('_', ' ').strip().upper()

class TermRate(Enum):
    SIMPLE = 'simple'
    CONTINUOUS = 'continuous'
    #DAILY = 'daily' # Forward rate formula is not implemented in ZeroCurve
    #WEEKLY = 'weekly' # Forward rate formula is not implemented in ZeroCurve
    #MONTHLY = 'monthly' # Forward rate formula is not implemented in ZeroCurve
    #QUARTERLY = 'quarterly' # Forward rate formula is not implemented in ZeroCurve
    #SEMIANNUAL = 'semiannual' # Forward rate formula is not implemented in ZeroCurve
    ANNUAL = 'annual'

    @property
    def display_name(self):
        return self.name.title()


class FXSmileInterpolationMethod(Enum):
    UNIVARIATE_SPLINE = 'univariate_spline'
    CUBIC_SPLINE = 'cubic_spline'
    HESTON_1993 = 'heston_1993'
    HESTON_CARR_MADAN_GAUSS_KRONROD_QUADRATURE = 'heston_carr_madan_gauss_kronrod_quadrature'
    HESTON_CARR_MADAN_FFT_W_SIMPSONS = 'heston_carr_madan_fft_w_simpsons'
    HESTON_LIPTON = 'heston_lipton'
    HESTON_COSINE = 'heston_cosine'

    @property
    def display_name(self):
        return self.name.title().replace('_', ' ')


class DeltaConvention(Enum):
    REGULAR_SPOT = 'regular_spot'
    REGULAR_FORWARD = 'regular_forward'
    PREMIUM_ADJUSTED_SPOT = 'premium_adjusted_spot'
    PREMIUM_ADJUSTED_FORWARD = 'premium_adjusted_forward'

    @property
    def display_name(self):
        return self.name.title().replace('_', ' ')


